'use strict';

Object.defineProperty(exports, "__esModule", {
	value: true
});
exports.treeOrder = exports.lookUpDiff = undefined;

var _slicedToArray2 = require('babel-runtime/helpers/slicedToArray');

var _slicedToArray3 = _interopRequireDefault(_slicedToArray2);

exports.differenceMatrix = differenceMatrix;
exports.agnes = agnes;

var _underscore_ext = require('./underscore_ext');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*
 * Store lower triangle as a flat array.
 * Have to compute (i, j) coords to a single coord.
 *
 * x x x x
 * 0 x x x
 * 1 2 x x
 * 3 4 5 x
 */
/*
[1, 0] -> 0 = 0 + 0
[2, 0] -> 1 = 1 + 0
[2, 1] -> 2 = 1 + 1
[3, 0] -> 3 = (2 + 1) + 0
[3, 1] -> 4 = (2 + 1) + 1
[3, 2] -> 5 = (2 + 1) + 2

At each coord i, we do the sum i  + i  - 1 + ... = (i * (i - 1)) / 2
Then add j
*/

var toCoord = function toCoord(i, j) {
	return i * (i - 1) / 2 + j;
};

function invCoord(d) {
	// solving d = i * (i - 1) / 2, for i
	// This looks expensive, but doesn't show up in the profiles.
	var i = Math.floor((1 + Math.sqrt(1 + 8 * d)) / 2),
	    offset = i * (i - 1) / 2;
	return { i: i, j: d - offset };
}

/* Should test storing the lower triangle as nested arrays, instead.
 * Would be less math, but slightly more memory. Would still have to
 * do the inverse coord computation when sorting, so not much less math.
 * [[x], [x x], [x x x], ...]
 */
function differenceMatrix(data, diff) {
	var len = data.length,
	    dlen = len * (len - 1) / 2,
	    d = new Array(dlen),
	    I = diff(data[0], data[0]); // compute single value for identity.

	/* compute lower triangle */
	for (var i = 1; i < len; ++i) {
		for (var j = 0; j < i; ++j) {
			d[toCoord(i, j)] = diff(data[i], data[j]);
		}
	}
	return [I, d];
}

var lookUpDiff = exports.lookUpDiff = function lookUpDiff(_ref) {
	var _ref2 = (0, _slicedToArray3.default)(_ref, 2),
	    I = _ref2[0],
	    d = _ref2[1];

	return function (i, j) {
		return i > j ? d[toCoord(i, j)] : i < j ? d[toCoord(j, i)] : I;
	};
};

function setP(nodes, p) {
	var _repeat = true;

	var _nodes, _p;

	while (_repeat) {
		_repeat = false;

		if (nodes.length === 0) {
			return;
		}
		var n = nodes.shift();
		n.p = p;
		if (n.left) {
			nodes.push(n.left);
			nodes.push(n.right);
		}
		_nodes = nodes;
		_p = p;
		nodes = _nodes;
		p = _p;
		_repeat = true;
		continue;
	}
}

// agnes method with simple linkage. Since simple linkage is the global minimum
// difference between any two points, we compute the distance matrix, sort it,
// and walk it in order to build clusters. I suspect this is n^2 log n, but not
// positive, as I haven't thought about the setP step, which updates a subtree
// with new cluster identity after a merge. Difference matrix is n^2. Sort
// is n^2 log n^2 = n^2 2 log n -> n^2 log n.
//
// Do we actually need to hold the full tree, just to extract the order?
// Maybe we can reduce the memory requirements?
function join(d, leaves, order, tree) {
	var n = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : leaves.length - 1;
	var oi = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : 0;
	var _repeat2 = true;

	var _d, _leaves, _order, _tree, _n, _oi;

	while (_repeat2) {
		_repeat2 = false;
		var k = order[oi],
		    _invCoord = invCoord(k),
		    i = _invCoord.i,
		    j = _invCoord.j,
		    li = leaves[i],
		    lj = leaves[j];

		if (li.p !== lj.p || li.p === undefined) {
			// not already in same cluster
			// place cluster according to dendsort MOLO rules, Sakai, et al
			tree = li.p ? lj.p ?
			// two clusters; put tightest on left
			li.p.d < lj.p.d ? { left: li.p, right: lj.p, d: li.p.d + d[k] } : { left: lj.p, right: li.p, d: lj.p.d + d[k] } :
			// cluster in i; put it on left
			{ left: li.p, right: lj, d: li.p.d + d[k] } : lj.p ?
			// cluster in j, put it on left
			{ left: lj.p, right: li, d: lj.p.d + d[k] } :
			// two leaves; put lower index (j) on left(?)
			{ left: lj, right: li, d: d[k] };

			setP([tree.left, tree.right], tree);

			if (n > 1) {
				_d = d;
				_leaves = leaves;
				_order = order;
				_tree = tree;
				_n = n - 1;
				_oi = oi + 1;
				d = _d;
				leaves = _leaves;
				order = _order;
				tree = _tree;
				n = _n;
				oi = _oi;
				_repeat2 = true;
				continue;
			} else {
				return tree;
			}
		}
		_d = d;
		_leaves = leaves;
		_order = order;
		_tree = tree;
		_n = n;
		_oi = oi + 1;
		d = _d;
		leaves = _leaves;
		order = _order;
		tree = _tree;
		n = _n;
		oi = _oi;
		_repeat2 = true;
		continue;
	}
}

function agnes(data, diff) {
	var _differenceMatrix = differenceMatrix(data, diff),
	    _differenceMatrix2 = (0, _slicedToArray3.default)(_differenceMatrix, 2),
	    d = _differenceMatrix2[1],
	    order = (0, _underscore_ext.range)(d.length).sort(function (i, j) {
		return d[i] - d[j];
	}),
	    leaves = (0, _underscore_ext.times)(data.length, function (i) {
		return { i: i, p: undefined };
	});

	return join(d, leaves, order);
}

function treeOrderN(nodes) {
	var acc = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
	var _repeat3 = true;

	var _nodes2, _acc;

	while (_repeat3) {
		_repeat3 = false;

		if (!nodes.length) {
			return acc;
		}
		var n = nodes.pop();
		if (!n.left) {
			acc.push(n.i);
		} else {
			nodes.push(n.right);
			nodes.push(n.left);
		}
		_nodes2 = nodes;
		_acc = acc;
		nodes = _nodes2;
		acc = _acc;
		_repeat3 = true;
		continue;
	}
}

var treeOrder = exports.treeOrder = function treeOrder(tree) {
	return treeOrderN([tree]);
};