'use strict';

var _extends2 = require('babel-runtime/helpers/extends');

var _extends3 = _interopRequireDefault(_extends2);

var _slicedToArray2 = require('babel-runtime/helpers/slicedToArray');

var _slicedToArray3 = _interopRequireDefault(_slicedToArray2);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * UCSC Xena Client
 * http://xena.ucsc.edu
 *
 * Zoom-related functions - transforms of drag and drag select events.
 */

// Dependencies
var _require = require('./exonLayout'),
    chromRangeFromScreen = _require.chromRangeFromScreen;

var multi = require('./multi');
var _ = require('./underscore_ext');

// Selectors
var directionSelector = function directionSelector(_ref) {
	var direction = _ref.direction;
	return direction;
};
var fieldTypeSelector = function fieldTypeSelector(_ref2) {
	var fieldType = _ref2.fieldType;
	return fieldType;
};

// Zoom interface
var zoom = {
	direction: multi(fieldTypeSelector),
	geneZoomed: multi(fieldTypeSelector),
	geneZoomLength: multi(fieldTypeSelector),
	maxGeneZoomLength: multi(fieldTypeSelector),
	overlay: multi(fieldTypeSelector),
	startEndPx: multi(directionSelector),
	supportsGeneZoom: multi(fieldTypeSelector),
	zoomTo: multi(fieldTypeSelector),
	zoomToLength: multi(fieldTypeSelector)
};

//
// Direction - either h for gene/horizontal zoom or v for samples/vertical zoom
//

var direction = function direction(start, end) {
	return Math.abs(start.x - end.x) > Math.abs(start.y - end.y) ? 'h' : 'v';
};

var directionWithGeneModel = function directionWithGeneModel(_ref3) {
	var start = _ref3.start,
	    end = _ref3.end,
	    zone = _ref3.zone;
	return zone === 'a' ? 'h' : direction(start, end);
};

var directionSamplesOnly = function directionSamplesOnly() {
	return 'v';
};

// @param {fieldType, start, end, zone}
zoom.direction.dflt = directionWithGeneModel;
zoom.direction.add('clinical', directionSamplesOnly);

//
// Start and end pixel values - use x start and end values for gene/horizontal zoom, use y start and end values for
// sample/vertical zoom
//

var flop = function flop(_ref4) {
	var start = _ref4.start,
	    end = _ref4.end;
	return start < end ? { start: start, end: end } : { start: end, end: start };
};

// @param {direction, start, end}
// -- start and end are pixel points in the target zone where the drag zoom event occurred (either samples or annotation)
zoom.startEndPx.add('h', function (_ref5) {
	var start = _ref5.start,
	    end = _ref5.end;
	return flop({ start: start.x, end: end.x });
});
zoom.startEndPx.add('v', function (_ref6) {
	var start = _ref6.start,
	    end = _ref6.end;
	return flop({ start: start.y, end: end.y });
});

//
// Overlay - return four points for drawing drag overlay: samples start, samples end, indicator start and indicator
// end (where indicator is either gene probe positioning for gene/horizontal zoom or "second sample zoom" positioning
// for samples/vertical zoom).
//

// Flip start and end chrom if layout is reversed
var flopChrom = function flopChrom(chromstart, chromend, reversed) {
	return reversed ? { start: chromstart, end: chromend } : { start: chromend, end: chromstart };
};

// True if start/end range falls within range of position, or spans position
var chromRangeInBin = function chromRangeInBin(_ref7, _ref8) {
	var _ref9 = (0, _slicedToArray3.default)(_ref7, 2),
	    chromstart = _ref9[0],
	    chromend = _ref9[1];

	var start = _ref8.start,
	    end = _ref8.end;
	return chromstart >= start && chromstart <= end || // start is in position
	chromend >= start && chromend <= end || // end is in position
	chromstart <= start && chromend >= end;
}; // start/end spans position

// Returns width of each subcolumn
var subcolumnWidth = function subcolumnWidth(columnWidth, subcolumns) {
	return columnWidth / subcolumns.length;
};

// Return pixel x coordinates for the specified subcolumn index
var subcolumnToPx = function subcolumnToPx(_ref10, i) {
	var layout = _ref10.layout,
	    position = _ref10.position;
	var pxLen = layout.pxLen,
	    width = subcolumnWidth(pxLen, position),
	    start = width * i,
	    end = start + width - 1;

	return { start: start, end: end };
};

// Return the set of subcolumn indices where the chromrange either falls into, or spans.
var overlapSubcolumnIndices = function overlapSubcolumnIndices(column, chromRange) {
	var layout = column.layout,
	    position = column.position,
	    reversed = layout.reversed;

	return position.map(function (_ref11, i) {
		var chromstart = _ref11.chromstart,
		    chromend = _ref11.chromend;
		return (0, _extends3.default)({ i: i }, flopChrom(chromstart, chromend, reversed));
	}).filter(function (pos) {
		return chromRangeInBin(chromRange, pos);
	}).map(function (_ref12) {
		var i = _ref12.i;
		return i;
	});
};

// Calculate subcolumn start and end pixel points from the specified annotation start and end pixel points.
var annotationPxToSamplesPx = function annotationPxToSamplesPx(column, startPx, endPx) {
	var layout = column.layout,
	    chromRange = chromRangeFromScreen(layout, startPx, endPx),
	    overlaps = overlapSubcolumnIndices(column, chromRange),
	    start = null,
	    end = null;

	if (overlaps.length) {
		var minIndex = _.min(overlaps),
		    maxIndex = _.max(overlaps);
		start = subcolumnToPx(column, minIndex).start;
		end = subcolumnToPx(column, maxIndex).end;
	}
	return { start: start, end: end };
};

// Calculate index of subcolumn from the specified pixel point within samples area.
var samplesPxToSubcolumn = function samplesPxToSubcolumn(columnWidth, subcolumns, px) {
	return Math.floor(px / subcolumnWidth(columnWidth, subcolumns));
};

// Calculate the indicator start and end points in annotation, from the samples start and end points.
var samplesPxToAnnotationPx = function samplesPxToAnnotationPx(column, direction, start, end) {
	var layout = column.layout,
	    position = column.position,
	    istart = null,
	    iend = null;

	if (direction === 'h') {
		// Find start and end subcolumns from pixel points
		var baseLen = layout.baseLen,
		    pxLen = layout.pxLen,
		    reversed = layout.reversed,
		    zoom = layout.zoom,
		    startIndex = samplesPxToSubcolumn(pxLen, position, start),
		    endIndex = samplesPxToSubcolumn(pxLen, position, end),
		    selectedPos = position.slice(startIndex, endIndex + 1),
		    minChrom = _.min(selectedPos.map(function (p) {
			return p.chromstart;
		})),
		    maxChrom = _.max(selectedPos.map(function (p) {
			return p.chromend;
		})),
		    chromRange = [minChrom, maxChrom],
		    subcolumns = position.filter(function (_ref13) {
			var chromstart = _ref13.chromstart,
			    chromend = _ref13.chromend;
			return chromRangeInBin(chromRange, { start: chromstart, end: chromend });
		}),
		    trueMinChrom = _.min(subcolumns.map(function (p) {
			return p.chromstart;
		})),
		    trueMaxChrom = _.max(subcolumns.map(function (p) {
			return p.chromend;
		})),
		    columnSize = pxLen / baseLen,
		    startPx = (trueMinChrom - zoom.start) * columnSize,
		    endPx = (trueMaxChrom - zoom.start) * columnSize;

		istart = _.max([Math.floor(reversed ? pxLen - endPx : startPx), 0]);
		iend = _.min([Math.floor(reversed ? pxLen - startPx : endPx), pxLen - 1]);
	}
	return { istart: istart, iend: iend };
};

var overlayFromAnnotation = function overlayFromAnnotation(_ref14) {
	var column = _ref14.column,
	    start = _ref14.start,
	    end = _ref14.end;

	var sStartEnd = annotationPxToSamplesPx(column, start, end);
	return { sstart: sStartEnd.start, send: sStartEnd.end, istart: start, iend: end };
};

// Calculate start and end points for zoom in samples area in column with subcolumns
var overlayFromSamples = function overlayFromSamples(_ref15) {
	var annotated = _ref15.annotated,
	    column = _ref15.column,
	    direction = _ref15.direction,
	    start = _ref15.start,
	    end = _ref15.end;
	return annotated ? (0, _extends3.default)({ sstart: start, send: end }, samplesPxToAnnotationPx(column, direction, start, end)) : { sstart: start, send: end, istart: start, iend: end };
};

var overlayWithSubcolumns = function overlayWithSubcolumns(params) {
	return params.zone === 'a' ? overlayFromAnnotation(params) : overlayFromSamples(params);
};

var overlayWithoutSubcolumns = function overlayWithoutSubcolumns(_ref16) {
	var start = _ref16.start,
	    end = _ref16.end;
	return { sstart: start, send: end, istart: start, iend: end };
};

// @param {annotated, column, direction, fieldType, start, end, zone}
// -- start and end are pixel points in the target zone where the drag zoom event occurred (either samples or annotation)
zoom.overlay.dflt = overlayWithoutSubcolumns;
zoom.overlay.add('geneProbes', overlayWithSubcolumns);

//
// Zoom - gene/horizontal zoom requires chromstart/chromend values for columns with corresponding gene model or subcolumn
// indices columns without a gene model. Samples/vertical zoom requires index and count.
//

var geneZoomWithGeneModel = function geneZoomWithGeneModel(_ref17) {
	var layout = _ref17.column.layout,
	    istart = _ref17.istart,
	    iend = _ref17.iend;

	var _chromRangeFromScreen = chromRangeFromScreen(layout, istart, iend),
	    _chromRangeFromScreen2 = (0, _slicedToArray3.default)(_chromRangeFromScreen, 2),
	    start = _chromRangeFromScreen2[0],
	    end = _chromRangeFromScreen2[1];

	return { start: start, end: end };
};

var geneZoomWithoutGeneModel = function geneZoomWithoutGeneModel(params) {
	var column = params.column,
	    sstart = params.sstart,
	    send = params.send,
	    width = column.width,
	    fieldList = column.fieldList,
	    fields = column.fields,
	    fieldType = column.fieldType,
	    xzoom = column.xzoom,
	    maxXZoom = column.maxXZoom,
	    zStart = xzoom ? xzoom.start : maxXZoom.start,
	    zEnd = xzoom ? xzoom.end : maxXZoom.end,
	    fieldsInZoomRange = fieldType === 'geneProbes' ? fields : fieldList.slice(zStart, zEnd + 1),
	    start = zStart + samplesPxToSubcolumn(width, fieldsInZoomRange, sstart),
	    end = zStart + samplesPxToSubcolumn(width, fieldsInZoomRange, send);

	return { start: start, end: end };
};

var samplesZoom = function samplesZoom(_ref18) {
	var yZoom = _ref18.yZoom,
	    sstart = _ref18.sstart,
	    send = _ref18.send;


	// project user mouse down and mouse up pixel positions on to the samples array.
	// sstart and send are the start and end pixels selected on the canvas
	// and will range from 0 to canvas height -1.

	// we want to find the first (lowest) sample index covered by sstart and last (highest) sample index
	// sample covered by send.

	var sampleCount = yZoom.count; // number of samples in the canvas
	var pixelCount = yZoom.height; // number of pixels in the canvas
	var samplesPerPx = sampleCount / pixelCount; // number of samples in a pixel
	var sampleIndexOffset = yZoom.index; // starting sample index in the canvas

	// to find the new starting index
	// sstart is the distance in pixels from the start of the canvas to the to start of the starting pixel
	// for example, the start of px 0 is 0 pixels from the start of the canvas, px 1 is 1 px from the start of the canvas.

	// sstart (pixels)  * samples/Pixel is samples.
	// take the floor of the samples to remove any fractional sample.
	// add the current starting index to get the new start index.
	var startIndex = Math.floor(sstart * samplesPerPx) + sampleIndexOffset;

	// to find the new ending index
	// add one to the send pixel index as we want the distance in pixels to the top of the pixel (the start of the next one)
	var endDistance = send + 1;
	var endIndex = Math.floor(endDistance * samplesPerPx) + sampleIndexOffset;

	// Accuracy of the floating point math can make endIndex greater than the last sample index.
	// Trim back if its over.
	var maxIndex = sampleCount + sampleIndexOffset - 1; // Determine the max index
	endIndex = endIndex > maxIndex ? maxIndex : endIndex; // Clip end index to maxIndex

	// Setup result
	var index = startIndex;
	var count = endIndex - startIndex + 1;
	return { index: index, count: count };
};

var zoomByDirection = function zoomByDirection(params) {
	return params.direction === 'v' ? samplesZoom(params) : params.annotated ? geneZoomWithGeneModel(params) : geneZoomWithoutGeneModel(params);
};

// @param {annotated, column, direction, fieldType, sstart, ssend, istart, iend, yZoom}
// -- sstart and ssend are pixel points in the samples zone
// -- istart and iend are pixel points in the annotation (indicator) zone
// -- yZoom is the current samples/vertical zoom state
zoom.zoomTo.dflt = zoomByDirection;
zoom.zoomTo.add('clinical', samplesZoom);

//
// Calculate the maximum possible gene zoom length
//

var maxGeneZoomRange = function maxGeneZoomRange(column) {
	return _.get(column.maxXZoom, ['end']) - _.get(column.maxXZoom, ['start']);
};

// x zoom is indices for columns without corresponding gene model, add 1 to get length
var maxGeneZoomRangeByIndices = function maxGeneZoomRangeByIndices(column) {
	return maxGeneZoomRange(column) + 1;
};

zoom.maxGeneZoomLength.dflt = maxGeneZoomRange;
['probes', 'genes'].forEach(function (fieldType) {
	zoom.maxGeneZoomLength.add(fieldType, maxGeneZoomRangeByIndices);
});

//
// Calculate current gene zoom length
//

var geneZoomRange = function geneZoomRange(column) {
	return _.get(column.xzoom, ['end']) - _.get(column.xzoom, ['start']);
};

// x zoom is indices for columns without corresponding gene model, add 1 to get length
var geneZoomRangeByIndices = function geneZoomRangeByIndices(column) {
	return geneZoomRange(column) + 1;
};

zoom.geneZoomLength.dflt = geneZoomRange;
['probes', 'genes'].forEach(function (fieldType) {
	zoom.geneZoomLength.add(fieldType, geneZoomRangeByIndices);
});

//
// Calculate length of zoom to
//

var zoomToLength = function zoomToLength(_ref19) {
	var zoomTo = _ref19.zoomTo;
	return _.get(zoomTo, ['end']) - _.get(zoomTo, ['start']);
};

// x zoom is indices for columns without corresponding gene model, add 1 to get length
var zoomToLengthByIndices = function zoomToLengthByIndices(params) {
	return zoomToLength(params) + 1;
};

zoom.zoomToLength.dflt = zoomToLength;
['probes', 'genes'].forEach(function (fieldType) {
	zoom.zoomToLength.add(fieldType, zoomToLengthByIndices);
});

//
// True if gene/horizontal zoom is possible for column type
//

var geneZoomSupported = function geneZoomSupported() {
	return true;
};
var geneZoomUnsupported = function geneZoomUnsupported() {
	return false;
};

zoom.supportsGeneZoom.dflt = geneZoomSupported;
zoom.supportsGeneZoom.add('clinical', geneZoomUnsupported);

//
// True if there is currently a gene zoom for the column
//

var geneZoomed = function geneZoomed(column) {
	return column.xzoom !== undefined && !(_.get(column.xzoom, ['start']) === _.get(column.maxXZoom, ['start']) && _.get(column.xzoom, ['end']) === _.get(column.maxXZoom, ['end']));
};

['geneProbes', 'segmented', 'mutation', 'probes', 'genes', 'SV', 'clinical'].forEach(function (fieldType) {
	zoom.geneZoomed.add(fieldType, geneZoomed);
});

module.exports = zoom;