'use strict';

var _set = require('babel-runtime/core-js/set');

var _set2 = _interopRequireDefault(_set);

var _slicedToArray2 = require('babel-runtime/helpers/slicedToArray');

var _slicedToArray3 = _interopRequireDefault(_slicedToArray2);

var _toConsumableArray2 = require('babel-runtime/helpers/toConsumableArray');

var _toConsumableArray3 = _interopRequireDefault(_toConsumableArray2);

var _clusterWorker = require('worker-loader!./cluster-worker');

var _clusterWorker2 = _interopRequireDefault(_clusterWorker);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Helper methods needed by multiple controllers.

var Rx = require('../rx');
var xenaQuery = require('../xenaQuery');
var _ = require('../underscore_ext');

var _require = require('./errors'),
    reifyErrors = _require.reifyErrors,
    collectResults = _require.collectResults;

var fetch = require('../fieldFetch');
var kmModel = require('../models/km');

var _require2 = require('../models/fieldSpec'),
    signatureField = _require2.signatureField;

var defaultServers = require('../defaultServers');
var publicServers = defaultServers.publicServers;

var gaEvents = require('../gaEvents');
// pick up signature fetch
require('../models/signatures');

var worker = new _clusterWorker2.default();

// XXX error handling? What do we do with errors in the worker?
var workerObs = Rx.Observable.fromEvent(worker, 'message').share();
var msgId = 0;

// sendMessage wraps worker messages in ajax-like observables, by assigning
// unique ids to each request, and waiting for a single response with the
// same id. The worker must echo the id in the response.
var sendMessage = function sendMessage(msg) {
	var id = msgId++;
	worker.postMessage({ msg: msg, id: id });
	return workerObs.filter(function (ev) {
		return ev.data.id === id;
	}).take(1).map(function (ev) {
		return ev.data.msg;
	});
};

// data is passed here due to a problem with evaluation order.
// We need data.avg, which is computed in the selector. So, we
// have to pass it in from a view, rather than take it from app state.
function fetchClustering(serverBus, state, id, data) {
	// maybe prune the data that we send?
	serverBus.next([['cluster-result', id], sendMessage(['cluster', data])]);
}

var datasetResults = function datasetResults(resps) {
	return collectResults(resps, function (servers) {
		return _.object(_.flatmap(servers, function (s) {
			return _.map(s.datasets, function (d) {
				return [d.dsID, d];
			});
		}));
	});
};

function datasetQuery(servers, cohort) {
	return Rx.Observable.zipArray(_.map(servers, function (server) {
		return reifyErrors(xenaQuery.datasetList(server, [cohort.name]).map(function (datasets) {
			return { server: server, datasets: datasets };
		}), { host: server });
	})).flatMap(datasetResults);
}

function fetchDatasets(serverBus, servers, cohort) {
	serverBus.next(['datasets', datasetQuery(servers, cohort)]);
}

var MAX_SAMPLES = 50 * 1000;

var allSamples = _.curry(function (cohort, max, server) {
	return xenaQuery.cohortSamples(server, cohort, max === Infinity ? null : max);
});

function unionOfGroup(gb) {
	return _.union.apply(_, (0, _toConsumableArray3.default)(_.map(gb, function (_ref) {
		var _ref2 = (0, _slicedToArray3.default)(_ref, 1),
		    v = _ref2[0];

		return v;
	})));
}

function logSampleSources(cohortResps) {
	var havingSamples = cohortResps.filter(function (_ref3) {
		var _ref4 = (0, _slicedToArray3.default)(_ref3, 1),
		    v = _ref4[0];

		return v.length > 0;
	}).map(function (_ref5) {
		var _ref6 = (0, _slicedToArray3.default)(_ref5, 3),
		    server = _ref6[2];

		return server;
	}),
	    types = new _set2.default(havingSamples.map(function (s) {
		return s === defaultServers.servers.localHub ? 'localhost' :
		// counting all ucsc-hosted hubs as public
		s.indexOf('.xenahubs.net') !== -1 ? 'public' : 'private';
	}));
	if (types.has('localhost')) {
		// user has samples on localhost hub
		gaEvents('hubs', 'localhost');
	}
	if (types.has('private')) {
		// user has samples on private hub that isn't localhost
		gaEvents('hubs', 'private');
	}
	if (types.size > 1 && (types.has('localhost') || types.has('private'))) {
		// user is joining samples across hubs, and not all of them
		// are xena public hubs.
		gaEvents('hubs', 'cohort join');
	}
}

// Performance of this is probably poor, esp. due to underscore's horrible
// n^2 set operations.
function cohortHasPrivateSamples(cohortResps) {
	var _$groupBy = _.groupBy(cohortResps, function (_ref7) {
		var _ref8 = (0, _slicedToArray3.default)(_ref7, 3),
		    server = _ref8[2];

		return _.contains(publicServers, server);
	}),
	    pub = _$groupBy['true'],
	    priv = _$groupBy['false'],
	    pubSamps = unionOfGroup(pub),
	    privSamps = unionOfGroup(priv);

	return _.difference(privSamps, pubSamps).length > 0;
}

function filterSamples(sampleFilter, samples) {
	return sampleFilter ? _.intersection(sampleFilter, samples) : samples;
}

// For the cohort, query all servers,
// return a stream per-cohort, each of which returns an event
// [cohort, [sample, ...]].
// By not combining them here, we can uniformly handle errors, below.
var cohortSamplesQuery = function cohortSamplesQuery(servers, max, _ref9) {
	var name = _ref9.name,
	    sampleFilter = _ref9.sampleFilter;
	return _.map(servers, allSamples(name, max)).map(function (resp, j) {
		return resp.map(function (samples) {
			return [filterSamples(sampleFilter, samples), samples.length >= max, servers[j]];
		});
	});
};

var collateSamples = _.curry(function (cohorts, max, resps) {
	var serverOver = _.any(resps, function (_ref10) {
		var _ref11 = (0, _slicedToArray3.default)(_ref10, 2),
		    over = _ref11[1];

		return over;
	}),
	    cohortSamples = unionOfGroup(resps || []).slice(0, max),
	    cohortOver = cohortSamples.length >= max,
	    hasPrivateSamples = cohortHasPrivateSamples(resps);
	logSampleSources(resps);
	return { samples: cohortSamples, over: serverOver || cohortOver, hasPrivateSamples: hasPrivateSamples };
});

// reifyErrors should be pass the server name, but in this expression we don't have it.
function samplesQuery(servers, cohort, max) {
	return Rx.Observable.zipArray(cohortSamplesQuery(servers, max, cohort).map(reifyErrors)).flatMap(function (resps) {
		return collectResults(resps, collateSamples(cohort, max));
	});
}

function fetchSamples(serverBus, servers, cohort, allowOverSamples) {
	serverBus.next(['samples', samplesQuery(servers, cohort, allowOverSamples ? Infinity : MAX_SAMPLES)]);
}

function fetchColumnData(serverBus, samples, id, settings) {
	//	if (Math.random() > 0.5) { // testing error handling
	serverBus.next([['widget-data', id], fetch(settings, samples)]);
	//	} else {
	//		serverBus.onNext([['widget-data', id], Rx.Observable.throw(new Error('Injected error'))]);
	//	}
}

function resetZoom(state) {
	var count = _.getIn(state, ['cohortSamples', 'length'], 0);
	return _.updateIn(state, ["zoom"], function (z) {
		return _.merge(z, { count: count, index: 0 });
	});
}

var setCohortRelatedFields = function setCohortRelatedFields(state, cohort) {
	return _.assoc(state, 'cohort', cohort, 'sampleSearch', '', 'hasPrivateSamples', false, 'cohortSamples', [], 'columns', {}, 'columnOrder', [], 'data', {}, 'survival', null, 'km', _.assoc(state.km, ['id'], null));
};

// This adds or overwrites a 'sample' column in the state.
// Called from setCohort, the column data will be fetched after
// the sample list returns from the server.
function addSampleColumn(state, width) {
	if (!_.get(state.cohort, ['name'])) {
		return state;
	}
	var field = signatureField('samples', {
		columnLabel: 'Sample ID',
		valueType: 'coded',
		signature: ['samples']
	}),
	    newOrder = _.has(state.columns, 'samples') ? state.columnOrder : [].concat((0, _toConsumableArray3.default)(state.columnOrder), ['samples']),
	    settings = _.assoc(field, 'width', Math.round(width == null ? 136 : width), 'user', _.pick(field, ['columnLabel', 'fieldLabel'])),
	    newState = _.assocIn(state, ['columns', 'samples'], settings, ['columnOrder'], newOrder);
	return _.assocIn(newState, ['data', 'samples', 'status'], 'loading');
}

var setWizardAndMode = function setWizardAndMode(state) {
	return _.assocIn(state, ['wizardMode'], true, ['mode'], 'heatmap');
};

var setCohort = _.curry(function (cohort, width, state) {
	return addSampleColumn(setWizardAndMode(resetZoom(setCohortRelatedFields(state, cohort))), width);
});

var userServers = function userServers(state) {
	return _.keys(state.servers).filter(function (h) {
		return state.servers[h].user;
	});
};

var fetchCohortData = function fetchCohortData(serverBus, state) {
	var user = userServers(state);
	if (state.cohort) {
		fetchSamples(serverBus, user, state.cohort, state.allowOverSamples);
	}
};

//
// survival fields
//

var hasSurvFields = function hasSurvFields(vars) {
	return !!_.some(_.values(kmModel.survivalOptions), function (option) {
		return vars[option.ev] && vars[option.tte] && vars[option.patient];
	});
};

var probeFieldSpec = function probeFieldSpec(_ref12) {
	var dsID = _ref12.dsID,
	    name = _ref12.name;
	return {
		dsID: dsID,
		fetchType: 'xena', // maybe take from dataset meta instead of hard-coded
		valueType: 'float',
		fieldType: 'probes',
		fields: [name]
	};
};

var codedFieldSpec = function codedFieldSpec(_ref13) {
	var dsID = _ref13.dsID,
	    name = _ref13.name;
	return {
		dsID: dsID,
		fetchType: 'xena', // maybe take from dataset meta instead of hard-coded
		valueType: 'coded',
		fieldType: 'clinical',
		fields: [name]
	};
};

function mapToObj(keys, fn) {
	return _.object(keys, _.map(keys, fn));
}

function survivalFields(cohortFeatures) {
	var vars = kmModel.pickSurvivalVars(cohortFeatures),
	    fields = {};

	if (hasSurvFields(vars)) {
		fields.patient = codedFieldSpec(vars.patient);

		_.values(kmModel.survivalOptions).forEach(function (option) {
			if (vars[option.ev] && vars[option.tte]) {
				fields[option.ev] = probeFieldSpec(vars[option.ev]);
				fields[option.tte] = probeFieldSpec(vars[option.tte]);
			}
		});

		if (_.has(fields, 'ev') && _.keys(fields).length > 3) {
			delete fields.ev;
			delete fields.tte;
		}
	}
	return fields;
}

// If field set has changed, re-fetch.
function fetchSurvival(serverBus, state) {
	var _Rx$Observable;

	var cohortFeatures = state.wizard.cohortFeatures,
	    _state$spreadsheet = state.spreadsheet,
	    cohort = _state$spreadsheet.cohort,
	    survival = _state$spreadsheet.survival,
	    cohortSamples = _state$spreadsheet.cohortSamples,
	    fields = survivalFields(cohortFeatures[cohort.name]),
	    survFields = _.keys(fields),
	    refetch = _.some(survFields, function (f) {
		return !_.isEqual(fields[f], _.getIn(survival, [f, 'field']));
	}),
	    queries = _.map(survFields, function (key) {
		return fetch(fields[key], cohortSamples);
	}),
	    collate = function collate(data) {
		return mapToObj(survFields, function (k, i) {
			return { field: fields[k], data: data[i] };
		});
	};

	refetch && serverBus.next(['km-survival-data', (_Rx$Observable = Rx.Observable).zipArray.apply(_Rx$Observable, (0, _toConsumableArray3.default)(queries)).map(collate)]);
}

module.exports = {
	fetchCohortData: fetchCohortData,
	fetchColumnData: fetchColumnData,
	fetchClustering: fetchClustering,
	fetchDatasets: fetchDatasets,
	fetchSamples: fetchSamples,
	fetchSurvival: fetchSurvival,
	resetZoom: resetZoom,
	setCohort: setCohort,
	userServers: userServers,
	datasetQuery: datasetQuery
};