'use strict';

Object.defineProperty(exports, "__esModule", {
	value: true
});

var _set = require('babel-runtime/core-js/set');

var _set2 = _interopRequireDefault(_set);

var _underscore_ext = require('../underscore_ext');

var _underscore_ext2 = _interopRequireDefault(_underscore_ext);

var _constants = require('./constants');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var MUTATION_BY_POS = _constants.DATA_TYPE.MUTATION_BY_POS,
    SEGMENTED_CN = _constants.DATA_TYPE.SEGMENTED_CN,
    GENOMIC = _constants.DATA_TYPE.GENOMIC,
    PHENOTYPE = _constants.DATA_TYPE.PHENOTYPE;


var HEADER_LENGTH_LIMIT = 255,
    MULTI_ERROR_LIMIT = 3;

var ERRORS = void 0,
    HEADER_ERRORS = void 0;

var getColumns = function getColumns(line) {
	return line.replace(/[\r\n]+$/, '').split(/\t/g);
};

var filterEmpty = function filterEmpty(arr) {
	return arr.filter(function (elem) {
		return !!elem;
	});
};

var getMutationHeaderRegExps = [{ regexp: /sample[ _]*(name|id)?/i, name: "'sample'" }, { regexp: /chr(om)?/i, name: "'chrom'" }, { regexp: /start/i, name: "'start'" }, { regexp: /end/i, name: "'end'" }, { regexp: /alt(ernate)?/i, name: "'alternate'" }, { regexp: /ref(erence)?/i, name: "'reference'" }];

var getSegmentedHeaderRegExps = [{ regexp: /sample[ _]*(name|id)?/i, name: "'sample'" }, { regexp: /chr(om)?/i, name: "'chrom'" }, { regexp: /start/i, name: "'start'" }, { regexp: /end/i, name: "'end'" }, { regexp: /value/i, name: "'value'" }];

var gatherErrorMessages = function gatherErrorMessages(state, dataType, errors, line, i) {
	errors.forEach(function (error) {
		error.getErrors(state, dataType, line, i);
	});
};

// Return function that error checks a line, keeping results (imperatively)
// in 'state'.
var getErrors = function getErrors(dataType, state) {
	var dataTypeErrors = ERRORS.filter(function (err) {
		return err.forType.some(function (dType) {
			return dType === dataType;
		});
	}),
	    dataTypeHeaderErrors = HEADER_ERRORS.filter(function (err) {
		return err.forType.some(function (dType) {
			return dType === dataType;
		});
	}),
	    i = 0;

	state.errors = [];
	state.warnings = [];
	state.snippets = [];

	return function (line) {
		//iterate through rules and populate state
		var lineCols = getColumns(line); // should this trim()? Or trim() earlier?
		if (i === 0) {
			state.rows = new _set2.default();
			state.headerLen = lineCols.length;
			gatherErrorMessages(state, dataType, dataTypeHeaderErrors, lineCols);
		} else {
			gatherErrorMessages(state, dataType, dataTypeErrors, lineCols, i);
		}

		i += 1;
	};
};

var hasColumn = function hasColumn(header, columnName, regexp) {
	if (!header.some(function (h) {
		return h.match(regexp);
	})) {
		return columnName;
	}
};

var hasSparseDataColumns = function hasSparseDataColumns(header, colRegExps) {
	return filterEmpty(colRegExps.map(function (r) {
		return hasColumn(header, r.name, r.regexp);
	}));
};

ERRORS = [{
	//HEADER COLUMN COUNT MISMATCH
	forType: [MUTATION_BY_POS, SEGMENTED_CN, GENOMIC, PHENOTYPE], // XXX for all types?
	getErrors: function getErrors(state, dataType, line, i) {
		var c = state.columnMismatch || 0;
		if (line.length !== state.headerLen && c < MULTI_ERROR_LIMIT) {
			state.errors.push('The number of headers doesn\'t match the number of columns on line ' + (i + 1) + '. \t\t\t\t\tPlease edit the file and reload.');
			state.columnMismatch = c + 1;
		}
	}
}, {
	//DUPLICATE ROW KEY
	forType: [GENOMIC, PHENOTYPE],
	getErrors: function getErrors(state, dataType, line /*, i*/) {
		if (!state.duplicateRows && state.rows.has(line[0])) {
			if (dataType === GENOMIC) {
				state.warnings.push('There are duplicate probe names in your file. An example is ' + line[0] + '. \t\t\t\t\t\tWe will load the first one and rename the duplicates.');
			} else {
				state.warnings.push('There are duplicate sample names in your file. An example is ' + line[0] + '. \t\t\t\t\t\tWe will load the first one and ignore all others.');
			}
			state.duplicateRows = true;
		} else {
			state.rows.add(_underscore_ext2.default.copyStr(line[0]));
		}
	}
}, {
	//GENOMIC MATRIX SHOULD ONLY HAS NUMBERS and NAs, NO OTHER VALUES SHOULD BE ALLOWED INCLUDING NULL
	forType: [GENOMIC],
	getErrors: function getErrors(state, dataType, line, i) {
		var badNumbers = line.filter(function (x, i) {
			return i !== 0 && isNaN(x) && x !== 'NA';
		});
		if (badNumbers.length) {
			state.errors.push('Only NUMBER (e.g. 1, 3.2) or \'NA\' is allowed in the genomic data. \t\t\t\t\tWe detected other values, such as \'' + badNumbers[0] + '\', on line ' + (i + 1) + '. \t\t\t\t\tPlease edit the file and reload.');
		}
	}
}];

HEADER_ERRORS = [{
	//HEADER LENGTHS
	forType: [GENOMIC, PHENOTYPE, MUTATION_BY_POS, SEGMENTED_CN],
	getErrors: function getErrors(state, dataType, line) {
		var message = function message(header) {
			return header.slice(0, 100) + '... is too long. Please limit to 255 characters';
		};

		var c = 0;
		for (var i = 0; i < line.length; i++) {
			var name = line[i];

			if (name.length > HEADER_LENGTH_LIMIT) {
				state.errors.push(message(name));
				c += 1;

				if (c >= MULTI_ERROR_LIMIT) {
					break;
				}
			}
		}
	}
}, { // XXX discard, or maybe check for quotes?
	//HEADER CHARACTERS
	forType: [GENOMIC, PHENOTYPE, MUTATION_BY_POS, SEGMENTED_CN],
	getErrors: function getErrors() /*state, line*/{
		// const message = (header) => `Headers can only have xyz. Please change ${header}`;
	}
}, {
	//HEADER EMPTY
	forType: [GENOMIC, PHENOTYPE, MUTATION_BY_POS, SEGMENTED_CN],
	getErrors: function getErrors(state, dataType, line) {
		if (line.some(function (h) {
			return h.trim() === '';
		})) {
			state.errors.push('One or more of your headers is blank. Please edit the file and reload.');
		}
	}
}, {
	//DUPLICATE HEADER KEYS
	forType: [GENOMIC, PHENOTYPE],
	getErrors: function getErrors(state, dataType, line) {
		var headers = new _set2.default(),
		    dup = _underscore_ext2.default.find(line, function (s) {
			if (headers.has(s)) {
				return true;
			}
			headers.add(s); // XXX note side-effecting predicate
		});
		if (dup) {
			if (dataType === GENOMIC) {
				state.warnings.push('There are duplicate samples in your file. An example is ' + dup + '. We will load the first one and ignore all others.');
			} else {
				state.warnings.push('There are duplicate phenotypes in your file. An example is ' + dup + '. We will load the first one and rename the duplicates.');
			}
		}
	}
}, {
	//SEGMENTED REQUIRED HEADERS
	forType: [SEGMENTED_CN],
	getErrors: function getErrors(state, dataType, line) {
		var missing = hasSparseDataColumns(line, getSegmentedHeaderRegExps);
		if (missing.length) {
			state.errors.push('For segmented copy number data we require 5 columns: \'sample\', \'chrom\', \'start\', \'end\', and \'value\'. You are missing ' + missing.join(', ') + '. Please edit your file and reload.');
			state.snippets.push({
				exampleLines: [['sample', 'chrom', 'start', 'end', 'value'], ['TCGA-OR-A5LT-01', 'chr1', '61735', '98588', '0.1842']]
			});
		}
	}
}, {
	//MUTATION REQUIRED HEADERS
	forType: [MUTATION_BY_POS],
	getErrors: function getErrors(state, dataType, line) {
		var missing = hasSparseDataColumns(line, getMutationHeaderRegExps);
		if (missing.length) {
			state.errors.push('For mutation data we require 6 columns: \'sample\', \'chrom\', \'start\', \'end\', \'reference\' and \'alternate\'. You are missing ' + missing.join(', ') + '. Please edit your file and reload.');
			state.snippets.push({
				exampleLines: [['sample', 'chrom', 'start', 'end', 'reference', 'alternate'], ['TCGA-PK-A5HB-01', 'chr1', '877831', '877831', 'T', 'C']]
			});
		}
	}
}];

exports.default = getErrors;