'use strict';

var _from = require('babel-runtime/core-js/array/from');

var _from2 = _interopRequireDefault(_from);

var _slicedToArray2 = require('babel-runtime/helpers/slicedToArray');

var _slicedToArray3 = _interopRequireDefault(_slicedToArray2);

var _set = require('babel-runtime/core-js/set');

var _set2 = _interopRequireDefault(_set);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Domain logic for segmented datasets.

var _ = require('../underscore_ext');
var widgets = require('../columnWidgets');
var xenaQuery = require('../xenaQuery');
var Rx = require('../rx');
var exonLayout = require('../exonLayout');
var intervalTree = require('static-interval-tree');

var _require = require('../layoutPlot'),
    pxTransformInterval = _require.pxTransformInterval;

var heatmapColors = require('../heatmapColors');
var parsePos = require('../parsePos');

function groupedLegend(colorMap, valsInData) {
	//eslint-disable-line no-unused-vars
	var inData = new _set2.default(valsInData),
	    groups = _.groupBy(_.filter(_.keys(colorMap), function (val) {
		return inData.has(val);
	}), function (k) {
		return colorMap[k];
	}),
	    colors = _.keys(groups);
	return {
		colors: colors,
		labels: _.map(colors, function (c) {
			return groups[c].join(', ');
		}),
		align: 'left'
	};
}

var exonPadding = {
	padTxStart: 3000,
	padTxEnd: 1000
};

var min = function min(x, y) {
	return x < y ? x : y;
};
var max = function max(x, y) {
	return x > y ? x : y;
};

// sum(len * value)/sum(len)
//
function segmentAverage(row, _ref) {
	var start = _ref.start,
	    end = _ref.end;

	var lengths = row.map(function (seg) {
		return min(seg.end, end) - max(seg.start, start) + 1;
	}),
	    totalLen = _.sum(lengths),
	    weightedSum = _.sum(row.map(function (seg, i) {
		return seg.value * lengths[i];
	}));
	return weightedSum / totalLen;
}

function cmp(column, data) {
	var sortVisible = _.get(column, 'sortVisible', true),
	    values = _.getIn(data, ['avg', sortVisible ? 'values' : 'geneValues', 0]);

	return values ? function (s1, s2) {
		return _.cmpNumberOrNull(values[s1], values[s2]);
	} : function () {
		return 0;
	};
}

var segmentedDataRange = xenaQuery.segmentedDataRange,
    refGeneRange = xenaQuery.refGeneRange;

//

// XXX Might want to optimize this before committing. We could mutate in-place
// without affecting anyone. This may be slow for large mutation datasets.
//
// Map sampleIDs to index into 'samples' array.

function mapSamples(samples, data) {
	var sampleMap = _.object(samples, _.range(samples.length));

	return _.updateIn(data, ['req', 'rows'], function (rows) {
		return _.map(rows, function (row) {
			return _.assoc(row, 'sample', sampleMap[row.sample]);
		});
	}, ['req', 'samplesInResp'], function (sIR) {
		return _.map(sIR, function (s) {
			return sampleMap[s];
		});
	});
}

function fetchChrom(_ref2, samples, pos) {
	var dsID = _ref2.dsID,
	    assembly = _ref2.assembly;

	var _ref3 = xenaQuery.refGene[assembly] || {},
	    name = _ref3.name,
	    host = _ref3.host;

	return refGeneRange(host, name, pos.chrom, pos.baseStart, pos.baseEnd).flatMap(function (refGene) {
		return segmentedDataRange(dsID, samples, pos.chrom, pos.baseStart, pos.baseEnd).map(function (req) {
			return mapSamples(samples, { req: req, refGene: refGene });
		});
	});
}

function fetchGene(_ref4, samples) {
	var dsID = _ref4.dsID,
	    fields = _ref4.fields,
	    assembly = _ref4.assembly;

	var _ref5 = xenaQuery.refGene[assembly] || {},
	    name = _ref5.name,
	    host = _ref5.host;

	return name ? xenaQuery.refGeneExons(host, name, fields).flatMap(function (refGene) {
		var coords = _.values(refGene)[0];
		if (!coords) {
			return Rx.Observable.of(null, Rx.Scheduler.asap);
		}
		var txStart = coords.txStart,
		    txEnd = coords.txEnd,
		    chrom = coords.chrom,
		    padTxStart = exonPadding.padTxStart,
		    padTxEnd = exonPadding.padTxEnd;

		return segmentedDataRange(dsID, samples, chrom, txStart - padTxStart, txEnd + padTxEnd).map(function (req) {
			return mapSamples(samples, { req: req, refGene: refGene });
		});
	}) : Rx.Observable.of(null, Rx.Scheduler.asap);
}

function fetch(column, cohortSamples) {
	var pos = parsePos(column.fields[0]),
	    method = pos ? fetchChrom : fetchGene;
	return method(column, cohortSamples, pos);
}

function findNodes(byPosition, layout, samples) {
	var sindex = _.object(samples, _.range(samples.length)),
	    minSize = function minSize(_ref6) {
		var _ref7 = (0, _slicedToArray3.default)(_ref6, 2),
		    s = _ref7[0],
		    e = _ref7[1];

		return [s, e - s < 1 ? s + 1 : e];
	};

	// _.uniq is something like O(n^2). Using ES6 Set, which should be more like O(n).
	var matches = new _set2.default(_.flatmap(layout.chrom, function (_ref8) {
		var _ref9 = (0, _slicedToArray3.default)(_ref8, 2),
		    start = _ref9[0],
		    end = _ref9[1];

		return intervalTree.matches(byPosition, { start: start, end: end });
	}));

	return (0, _from2.default)(_.iterable.map(matches, function (v) {
		var _minSize = minSize(pxTransformInterval(layout, [v.start, v.end])),
		    _minSize2 = (0, _slicedToArray3.default)(_minSize, 2),
		    xStart = _minSize2[0],
		    xEnd = _minSize2[1];

		return {
			xStart: xStart,
			xEnd: xEnd,
			y: sindex[v.segment.sample],
			value: v.segment.value,
			data: v.segment
		};
	}));
}
var swapIf = function swapIf(strand, _ref10) {
	var _ref11 = (0, _slicedToArray3.default)(_ref10, 2),
	    x = _ref11[0],
	    y = _ref11[1];

	return strand === '-' ? [y, x] : [x, y];
};

function defaultXZoom(pos, refGene) {
	if (pos) {
		// User supplied chrom position
		return {
			start: pos.baseStart,
			end: pos.baseEnd
		};
	}

	var txStart = refGene.txStart,
	    txEnd = refGene.txEnd,
	    strand = refGene.strand,
	    padTxStart = exonPadding.padTxStart,
	    padTxEnd = exonPadding.padTxEnd,
	    _swapIf = swapIf(strand, [padTxStart, padTxEnd]),
	    _swapIf2 = (0, _slicedToArray3.default)(_swapIf, 2),
	    startPad = _swapIf2[0],
	    endPad = _swapIf2[1];

	return {
		start: txStart - startPad,
		end: txEnd + endPad
	};
}

function dataToDisplay(column, vizSettings, data, sortedSamples, index) {
	var pos = parsePos(column.fields[0]);
	if (_.isEmpty(data) || _.isEmpty(data.req) || !pos && _.isEmpty(data.refGene)) {
		return {
			color: ['no-data']
		};
	}
	var refGeneObj = _.values(data.refGene)[0],
	    maxXZoom = defaultXZoom(pos, refGeneObj),
	    dataset = column.dataset,
	    width = column.width,
	    _column$showIntrons = column.showIntrons,
	    showIntrons = _column$showIntrons === undefined ? false : _column$showIntrons,
	    _column$xzoom = column.xzoom,
	    xzoom = _column$xzoom === undefined ? maxXZoom : _column$xzoom,
	    createLayout = pos ? exonLayout.chromLayout : showIntrons ? exonLayout.intronLayout : exonLayout.layout,
	    layout = createLayout(refGeneObj, width, xzoom, pos),
	    nodes = findNodes(index.byPosition, layout, sortedSamples),
	    color = heatmapColors.colorSpec(column, vizSettings, null, _.getIn(data, ['avg', 'geneValues', 0])),
	    units = [_.get(dataset, 'unit')];


	return {
		layout: layout,
		nodes: nodes,
		maxXZoom: maxXZoom,
		color: color,
		units: units
	};
}

function index(fieldType, data) {
	if (!_.get(data, 'req')) {
		return { bySample: {} };
	}

	var _data$req = data.req,
	    rows = _data$req.rows,
	    samplesInResp = _data$req.samplesInResp,
	    bySample = _.groupBy(rows, 'sample'),
	    empty = []; // use a single empty object.

	rows = rows.map(function (row) {
		var start = row.start,
		    end = row.end;


		return {
			start: start,
			end: end,
			segment: row
		};
	});

	return {
		byPosition: intervalTree.index(rows),
		bySample: _.object(samplesInResp, samplesInResp.map(function (s) {
			return bySample[s] || empty;
		}))
	};
}

///////////////////////////////////
// download

function makeRow(fields, sampleGroup, row) {
	var fieldValue = void 0;
	if (_.isArray(sampleGroup) && sampleGroup.length === 0) {
		fieldValue = 'no variant';
	}
	if (_.isEmpty(sampleGroup)) {
		sampleGroup = [row];
	}
	return _.flatmap(sampleGroup, function (row) {
		return _.map(fields, function (f) {
			return row && row[f] || fieldValue;
		});
	});
}

function getRowFields(rows, sampleGroups) {
	if (_.isEmpty(sampleGroups)) {
		return []; // When no samples exist
	} else if (!_.isEmpty(rows)) {
		return _.keys(rows[0]); // When samples have mutation(s)
	} else {
		return ['sample', 'result']; // default fields for mutation-less columns
	}
}

function formatSamples(sampleFormat, rows) {
	return rows ? _.map(rows, function (r) {
		return _.updateIn(r, ['sample'], sampleFormat);
	}) : rows;
}

function download(_ref12) {
	var column = _ref12.column,
	    data = _ref12.data,
	    samples = _ref12.samples,
	    sampleFormat = _ref12.sampleFormat;

	var geneAverages = _.getIn(data, ['avg', 'geneValues', 0]),
	    columnLabel = column.user.fieldLabel || column.fieldLabel;

	return [['sample', columnLabel + ' (average)'], samples.map(function (sample) {
		return [sampleFormat(sample), geneAverages[sample]];
	})];
}

function downloadOneSampleOneRow(_ref13) {
	var rows = _ref13.data.req.rows,
	    samples = _ref13.samples,
	    index = _ref13.index,
	    sampleFormat = _ref13.sampleFormat;

	var groupedSamples = _.getIn(index, ['bySample']) || [],
	    rowFields = getRowFields(rows, groupedSamples),
	    allRows = _.map(samples, function (sId) {
		var alternateRow = { sample: sampleFormat(sId) }; // only used for mutation-less samples
		return makeRow(rowFields, formatSamples(sampleFormat, groupedSamples[sId]), alternateRow);
	});
	return {
		type: "txt",
		downloadData: [rowFields, allRows]
	};
}

var avgOrNull = function avgOrNull(rows, xzoom, zero) {
	return rows ? _.isEmpty(rows) ? zero : segmentAverage(rows, xzoom) : null;
};

function avgSegWithZoom(count, samplesInResp, zero, byPosition, zoom) {
	var matches = _.pluck(intervalTree.matches(byPosition, zoom), 'segment'),
	    perSamp = _.groupBy(matches, 'sample'),
	    perSamplesInResp = _.object(samplesInResp, samplesInResp.map(function (i) {
		return perSamp[i] || [];
	}));

	return _.times(count, function (i) {
		return avgOrNull(perSamplesInResp[i], zoom, zero);
	});
}

function chromLimits(pos) {
	return {
		start: pos.baseStart,
		end: pos.baseEnd
	};
}

function geneLimits(refGene) {
	var gene = _.values(refGene)[0];
	return {
		start: gene.txStart,
		end: gene.txEnd
	};
}

// Average segments, clipping to zoom or the gene boundaries, whichever is smaller.
function averageSegments(column, data, count, index) {
	var pos = parsePos(column.fields[0]);
	if (!_.get(data, 'req') || !(pos || _.values(data.refGene).length)) {
		return null;
	}
	var limits = pos ? chromLimits(pos) : geneLimits(data.refGene),
	    xzoom = {
		start: max(limits.start, _.getIn(column, ['xzoom', 'start'], -Infinity)),
		end: min(limits.end, _.getIn(column, ['xzoom', 'end'], Infinity))
	},
	    samplesInResp = _.getIn(data, ['req', 'samplesInResp']),
	    zeroOrigin = _.getIn(column, ['vizSettings', 'origin'], undefined),
	    norm = _.getIn(column, ['vizSettings', 'colNormalization']),
	    zeroNorm = norm === 'normal2' ? 2 : norm === 'none' ? 0 : undefined,
	    defaultNorm = _.getIn(column, ['defaultNormalization']),
	    zeroDefaultNorm = defaultNorm === 'normal2' ? 2 : defaultNorm === 'none' ? 0 : undefined,
	    zero = zeroOrigin !== undefined ? zeroOrigin : zeroNorm !== undefined ? zeroNorm : zeroDefaultNorm !== undefined ? zeroDefaultNorm : 0,
	    values = [avgSegWithZoom(count, samplesInResp, zero, index.byPosition, xzoom)],
	    geneValues = [avgSegWithZoom(count, samplesInResp, zero, index.byPosition, { start: limits.start, end: limits.end })];

	return {
		avg: {
			values: values,
			// re-calculating this isn't really necessary. We could move it earlier, like in the 'index'
			// selector, since it doesn't depend on zoom.
			geneValues: geneValues
			// XXX do we need the mean? Only for a normalized view. Are we doing that?
			//			mean: values.map(_.meannull)
		}
	};
}

widgets.cmp.add('segmented', cmp);
widgets.index.add('segmented', index);
widgets.transform.add('segmented', dataToDisplay);
widgets.avg.add('segmented', averageSegments);
widgets.download.add('segmented', download);
widgets.specialDownload.add('segmented', downloadOneSampleOneRow);

module.exports = {
	fetch: fetch
};